from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Any

from predibase._errors import PredibaseError
from predibase.resource.llm.interface import _LLMDeployment
from predibase.resources.deployment import DedicatedLLMDeployment, ServerlessLLMDeployment
from predibase.util import log_info

if TYPE_CHECKING:
    from predibase import Predibase
    from predibase.resources.model import FinetunedLLMAdapter
    from predibase.resource.llm.response import GeneratedResponse


class Completions:
    def __init__(self, client: Predibase):
        self._client = client

    def create(
        self,
        data: str | Dict[str, Any],
        *,
        adapter: FinetunedLLMAdapter | None = None,  # TODO: support passing model URI
        deployment: str | DedicatedLLMDeployment | ServerlessLLMDeployment = "auto",
        temperature: float = 0.1,
        new_tokens: int = 128,
        bypass_system_prompt: bool = False,
    ) -> GeneratedResponse:
        """
        Creates a new text completion using the specified adapter and/or deployed LLM.

        :param data: Either a string prompt to pass to the LLM, or a dictionary mapping variable names to values. These
            variable names must appear in the prompt template with which the adapter was fine-tuned, and the mapped
            values will be interpolated into the adapter's prompt template.
        :param adapter: (Optional) The fine-tuned adapter to use for this prompt completion. If not specified, then
            a `deployment` to prompt must be provided.
        :param deployment: (Optional) The deployed LLM instance to prompt. If `adapter` is specified but this field is
            not, a deployment compatible with the selected adapter will be automatically chosen. If `adapter` is not
            specified, then this field must be non-empty.
        :param temperature: Higher temperature settings provide more diversity in LLM responses, while a lower
            temperature will result in more deterministic responses.
        :param new_tokens: The maximum number of output tokens the LLM should generate.
        :param bypass_system_prompt: If True, ignore any default prompt attached to the deployed LLM.
        :return: A GeneratedResponse object.
        """
        if isinstance(deployment, str) and deployment == "auto":
            # TODO: allow passing deployment as string

            if adapter is None:
                raise PredibaseError(f"the target deployment cannot be 'auto' if no adapter is provided")

            deployments = self._client._session.get_json("/llms")
            llms = [_LLMDeployment.from_dict(d) for d in deployments]
            compatible_serverless_llms = [
                llm
                for llm in llms
                if self._adapter_compatible(llm.model_name, adapter.base_model_name) and llm.is_shared
            ]
            compatible_dedicated_llms = [
                llm
                for llm in llms
                if self._adapter_compatible(llm.model_name, adapter.base_model_name) and not llm.is_shared
            ]

            if compatible_serverless_llms:
                deployment = ServerlessLLMDeployment(self._client, compatible_serverless_llms[0].name)
            elif compatible_dedicated_llms:
                deployment = DedicatedLLMDeployment(self._client, compatible_dedicated_llms[0].name)
            else:
                raise PredibaseError(
                    f"no compatible deployments found for adapter's base model {adapter.base_model_name}"
                )

            log_info(
                f"Automatically selected deployment {deployment.name} from those compatible with adapter's base model "
                f"{adapter.base_model_name}"
            )

        # Falling back to the underlying prompt/generate implementation on LLMDeployment for now in order
        # to better pick up expected changes for introducing LoRAX client.
        # TODO: cleaner split once we have more confidence in the new version.
        dep = deployment._refresh()
        if adapter:
            dep = dep.with_adapter(adapter._model)

        return dep.prompt(
            data,
            temperature=temperature,
            max_new_tokens=new_tokens,
            bypass_system_prompt=bypass_system_prompt,
        )

    @staticmethod
    def _adapter_compatible(deployment_model_name, adapter_model_name) -> bool:
        """
        Shortcut helper method that checks for adapter compatibility using some string manipulation
        on the deployment and adapter model names.
        """

        # TODO(PUX-2100): replace this with a more robust API call once it's available.

        def preprocess_name(name: str) -> str:
            ret = name.lower().split("/")[-1]
            if "-hf" in ret:
                ret = ret[: ret.index("-hf")]
            return ret

        deployment_model = preprocess_name(deployment_model_name)
        adapter_model = preprocess_name(adapter_model_name)
        if "-v0.1" in adapter_model:
            adapter_model = adapter_model[: adapter_model.index("-v0.1")]

        return adapter_model in deployment_model



from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_json import dataclass_json, LetterCase


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DataField:
    name: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ConnectionObject:
    """Represents an object in the connection that can be turned into a dataset."""

    name: str
    fields: Optional[List[DataField]] = field(default=None)

    def __repr__(self):
        s = self.name
        if self.fields is not None:
            s += " (" + ", ".join([f.name for f in self.fields]) + ")"
        return s



from typing import Any

import typer
from rich import print

from predibase.util.settings import load_global_settings, load_local_settings, load_settings, save_global_settings

app = typer.Typer(no_args_is_help=True)


@app.command(help="Show global settings")
def show_global():
    settings = load_global_settings()
    print(settings)


@app.command(help="Show local settings")
def show_local():
    settings = load_local_settings()
    print(settings)


@app.command(help="Show local and global settings")
def show_all():
    settings = load_settings()
    print(settings)


@app.command(help="Set default repository name")
def set_repo(repository_name: str = typer.Argument(..., help="The model repository name")):
    _set_setting("repo", repository_name)


@app.command(help="Set default engine name")
def set_engine(engine_name: str = typer.Argument(..., help="The engine name")):
    _set_setting("engine", engine_name)


@app.command(help="Set api token")
def set_api_token(token: str = typer.Argument(..., help="The api token")):
    _set_setting("token", token)


@app.command(help="Set default endpoint url")
def set_endpoint(endpoint: str = typer.Argument(..., help="The endpoint url")):
    _set_setting("endpoint", endpoint)


def _set_setting(setting: str, value: Any):
    settings = load_global_settings()
    settings[setting] = value
    save_global_settings(settings)



from typing import List, Optional, Union

import pandas as pd

from predibase.pql.api import ServerResponseError
from predibase.resource.connection import Connection
from predibase.resource.connection_properties import (
    ADLSProperties,
    BigQueryProperties,
    ConnectionProperties,
    ConnectionType,
    DatabaseProperties,
    DatabricksProperties,
    GCSProperties,
    S3Properties,
    SnowflakeProperties,
)
from predibase.util import log_info


class ConnectionMixin:
    def list_connections(
        self,
        name: str = None,
        limit: int = 9999999,
        df: bool = False,
    ) -> Union[pd.DataFrame, List[Connection]]:
        """
        :param name: filter by connection name
        :param limit: max number of connections to return
        :param df: Whether to return pandas dataframe or list of objects
        :return: Pandas dataframe or list of connections
        """
        endpoint = f"/connections?limit={limit}"
        if name:
            endpoint += "&searchKeys[]=name&searchVals[]=" + name
        resp = self.session.get_json(endpoint)
        connections = [x for x in resp["connections"]]
        if df:
            return pd.DataFrame(connections)
        return [Connection.from_dict({"session": self.session, **c}) for c in connections]

    def create_connection(
        self,
        name: str,
        connection_type: ConnectionType,
        properties: ConnectionProperties,
        exists_ok: bool = False,
    ) -> Connection:
        if self.session.is_plan_expired():
            raise PermissionError(
                "Connecting data is locked for expired plans. Contact us to upgrade.",
            )

        conn_with_secret = {
            "connection": {
                "name": name,
                "type": connection_type.value,
            },
            "secret": {
                "secret": properties.to_dict(),
            },
        }

        try:
            resp = self.session.post_json("/connections", conn_with_secret)
            return Connection.from_dict({"session": self.session, **resp})
        except ServerResponseError as e:
            if exists_ok and e.code == 400:
                log_info(f"Connection {name} already exists. exists_ok=True, so ignoring.")
                return self.get_connection(name)
            else:
                raise e

    def create_connection_postgres(
        self,
        name: str,
        address: str,
        database: str,
        username: str,
        password: str,
        exists_ok: bool = False,
    ) -> Connection:
        properties = DatabaseProperties(address=address, database=database, username=username, password=password)
        return self.create_connection(name, ConnectionType.POSTGRESQL, properties, exists_ok)

    def create_connection_s3(self, name: str, key: str, secret: str, exists_ok: bool = False) -> Connection:
        """Create connection with Amazon S3 storage.

        :param name: Name of connection.
        :param key: Amazon access key ID.
        :param secret: Amazon secret access key.
        :param exists_ok: If True, ignore if connection already exists.
        :return: A new connection.
        """
        properties = S3Properties(key=key, secret=secret)
        return self.create_connection(name, ConnectionType.S3, properties, exists_ok)

    def create_connection_mysql(
        self,
        name: str,
        address: str,
        database: str,
        username: str,
        password: str,
        exists_ok: bool = False,
    ) -> Connection:
        properties = DatabaseProperties(address=address, database=database, username=username, password=password)
        return self.create_connection(name, ConnectionType.MYSQL, properties, exists_ok)

    def create_connection_redshift(
        self,
        name: str,
        address: str,
        database: str,
        username: str,
        password: str,
        exists_ok: bool = False,
    ) -> Connection:
        properties = DatabaseProperties(address=address, database=database, username=username, password=password)
        return self.create_connection(name, ConnectionType.REDSHIFT, properties, exists_ok)

    def create_connection_snowflake(
        self,
        name: str,
        user: str,
        password: str,
        account: str,
        database: str,
        warehouse: str,
        schema: str,
        exists_ok: bool = False,
    ) -> Connection:
        properties = SnowflakeProperties(
            user=user,
            password=password,
            account=account,
            database=database,
            warehouse=warehouse,
            schema=schema,
        )
        return self.create_connection(name, ConnectionType.SNOWFLAKE, properties, exists_ok)

    def create_connection_adls(
        self,
        name: str,
        account_key: str,
        connection_string: str,
        exists_ok: bool = False,
    ) -> Connection:
        properties = ADLSProperties(account_key=account_key, connection_string=connection_string)
        return self.create_connection(name, ConnectionType.ADLS, properties, exists_ok)

    def create_connection_gcs(self, name: str, token: str, exists_ok: bool = False) -> Connection:
        properties = GCSProperties(token=token)
        return self.create_connection(name, ConnectionType.GCS, properties, exists_ok)

    def create_connection_bigquery(
        self,
        name: str,
        project: str,
        dataset: str,
        token: str,
        exists_ok: bool = False,
    ) -> Connection:
        properties = BigQueryProperties(project=project, dataset=dataset, token=token)
        return self.create_connection(name, ConnectionType.BIGQUERY, properties, exists_ok)

    def create_connection_databricks(
        self,
        name: str,
        access_token: str,
        server_host_name: str,
        http_path: str,
        schema: str = None,
        exists_ok: bool = False,
    ) -> Connection:
        properties = DatabricksProperties(
            access_token=access_token,
            server_host_name=server_host_name,
            http_path=http_path,
            schema_name=schema,
        )
        return self.create_connection(name, ConnectionType.DATABRICKS, properties, exists_ok)

    def get_connection(self, connection_name: Optional[str] = None, connection_id: Optional[int] = None) -> Connection:
        if not (connection_name or connection_id):
            raise ValueError("Must provide either connection_name or connection_id")

        endpoint = f"/connections/name/{connection_name}" if connection_name else f"/connections/{connection_id}"
        resp = self.session.get_json(endpoint)
        return Connection.from_dict({"session": self.session, **resp})

    def delete_connection(self, connection_name: str = None, connection_id: int = None):
        if connection_name and connection_id:
            raise ValueError("Cannot provide both connection_name and connection_id")
        if not (connection_name or connection_id):
            raise ValueError("Must provide either connection_name or connection_id")

        if connection_id:
            return self.session.delete_json("/connections", {"id": connection_id})
        return self.session.delete_json(f"/connections/name/{connection_name}")



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict, Any

from predibase._errors import PredibaseError
from predibase.resource.model import ModelFuture
from predibase.resource_util import build_model

if TYPE_CHECKING:
    from predibase import Predibase


logger = logging.getLogger(__name__)


class Model:
    pass


class LLM(Model):
    pass


class PretrainedLLM(LLM):
    pass


class PretrainedHuggingFaceLLM(PretrainedLLM):
    def __init__(self, name: str):
        # TODO: check the validity of the provided name via HuggingFace
        self.name = name
        if self.name.startswith("hf://"):
            self.name = self.name[len("hf://") :]


class FinetunedLLM(LLM):
    pass


class FinetunedLLMAdapter(FinetunedLLM):
    # TODO: accept version instead
    def __init__(self, client: Predibase, repo_name: str, model_id: int):
        self._model_id = model_id
        self._client = client
        self._session = client._session  # Directly using the session in the short term as we transition to v2.
        self._repo_name = repo_name

        self._refresh()

    def _refresh(self):
        endpoint = f"/models/version/{self._model_id}"
        resp = self._session.get_json(endpoint)
        self._model = build_model(resp["modelVersion"], self._session)

    @property
    def uri(self) -> str:
        return f"pb://models/{self._repo_name}/{self._model_id}"

    @property
    def train_config(self) -> Dict:
        return self._model.config

    @property
    def base_model_name(self) -> str:
        return self._model.llm_base_model_name  # Should never be None for an adapter.

    def wait_ready(self, show_tensorboard: bool = False) -> FinetunedLLMAdapter:
        self._refresh()
        status = self._model.status

        if status in ("ready", "deploying", "deployed", "undeploying"):
            return self

        if status == "canceled":
            raise PredibaseError(f"model {self._model_id} was canceled")

        if status == "failed":
            raise PredibaseError(f"model {self._model_id} failed to train: {self._model.error_text}")

        _ = ModelFuture(self._model_id, self._session).get(launch_tensorboard=show_tensorboard)
        return self



from __future__ import annotations

import mimetypes
import os
from os import PathLike
from typing import TYPE_CHECKING

import requests
from google.protobuf.json_format import MessageToJson
from predibase_api.artifacts.v1.artifacts_pb2 import (
    PresignedUrlForUploadRequest,
    ARTIFACT_TYPE_DATASET,
    PresignedUrlForUploadResponse,
    RegisterUploadedFileAsDatasetRequest,
)

from predibase._errors import PredibaseError
from predibase.resource.connection import get_dataset
from predibase.resource.dataset import Dataset

if TYPE_CHECKING:
    from predibase.pql import Session


class Datasets:
    def __init__(self, session: Session):
        self._session = session

    def get(self, resource_id: str) -> Dataset:
        if resource_id.startswith("pb://datasets/"):
            resource_id = resource_id[len("pb://datasets/"):]

        tokens = resource_id.split("/")
        if len(tokens) == 1:
            connection_name = None
            dataset_name = tokens[0]
        elif len(tokens) == 2:
            connection_name, dataset_name = tokens
        else:
            raise PredibaseError(f"unable to parse resource ID {resource_id}")

        return get_dataset(self._session, connection_name=connection_name, dataset_name=dataset_name)

    def from_file(self, file_path: PathLike, *, name: str | None = None) -> Dataset:
        with open(file_path, "rb") as f:
            file_name = os.path.basename(file_path)
            mime_type = mimetypes.guess_type(file_path)[0] or "text/plain"

            # Get presigned url for upload
            presigned_url_req = MessageToJson(
                PresignedUrlForUploadRequest(
                    mime_type=mime_type,
                    expiry_seconds=3600,
                    artifact_type=ARTIFACT_TYPE_DATASET,
                ),
                preserving_proto_field_name=True,
            )
            resp = PresignedUrlForUploadResponse(
                **self._session.post(
                    "/datasets/get_presigned_url",
                    data=presigned_url_req,
                ),
            )

            # Unpack response data
            presigned_url = resp.url
            object_name = resp.object_name
            required_headers = resp.required_headers

            # Get required headers for presigned url upload
            headers = {"Content-Type": mime_type}
            for k, v in required_headers:
                headers[k] = v

            # Upload file to blob storage with pre-signed url
            requests.put(presigned_url, data=f, headers=headers).raise_for_status()

            # Register uploaded file as dataset
            register_dataset_req = MessageToJson(
                RegisterUploadedFileAsDatasetRequest(
                    dataset_name=name,
                    object_name=object_name,
                    file_name=file_name,
                ),
                preserving_proto_field_name=True,
            )
            resp = self._session.post("/datasets/register_uploaded_file", data=register_dataset_req)

            dataset_id = resp["id"]
            endpoint = f"/datasets/{dataset_id}"
            resp = self._session.wait_for_dataset(endpoint, until_fully_connected=True)

            from predibase.resource_util import build_dataset

            return build_dataset(resp, self._session)
